;; -*-scheme-*-
;; by  Richard -Gilligan- Uschold 
;;
;; This prints Tax related accounts and exports TXF files for import to
;; TaxCut, TurboTax, etc.
;;
;; For this to work, the user has to segregate taxable and not taxable
;; income to different accounts, as well as deductible and non
;; deductible expenses.
;;
;; The user selects the accounts(s) to be printed, if none, all are checked.
;; Automatically prints up to 15 sub-account levels below selected
;; account.  Accounts below that are not printed. If you really need
;; more levels, change the MAX_LEVELS constant
;;
;; Optionally, does NOT print accounts with $0.00 values.  Prints data
;; between the From and To dates.  Optional alternate periods:
;; "Last Year", "1st Est Tax Quarter", ... "4th Est Tax Quarter"
;; "Last Yr Est Tax Qtr", ... "Last Yr Est Tax Qtr"
;; Estimated Tax Quarters: Dec 31, Mar 31, Jun 30, Aug 31)
;; Optionally prints brief or full account names
;;
;; NOTE: setting of specific dates is squirly! and seems to be
;; current-date dependant!  Actually, time of day dependant!  Just
;; after midnight gives diffenent dates than just before!  Referencing
;; all times to noon seems to fix this.  Subtracting 1 year sometimes
;; subtracts 2!  see "(to-value"

(gnc:support "report/taxtxf.scm")
(gnc:depend "text-export.scm")
(gnc:depend "report-utilities.scm")
(gnc:depend "options.scm")
(gnc:depend "date-utilities.scm")
(gnc:depend "report/txf-export.scm")
(gnc:depend "report/txf-export-help.scm")

(define (make-level-collector num-levels)
  (let ((level-collector (make-vector num-levels)))
    (do ((i 0 (+ i 1)))
        ((= i num-levels) i)
      (vector-set! level-collector i (gnc:make-stats-collector)))
    level-collector))

;; Just a private scope.
(let* ((MAX-LEVELS 16)			; Maximum Account Levels
       (levelx-collector (make-level-collector MAX-LEVELS))
       (today (gnc:timepair-canonical-day-time
	       (cons (current-time) 0)))
       (bdtm (gnc:timepair->date today))
       (junk (set-tm:mday bdtm 16))	; 16
       (junk (set-tm:mon bdtm 3))	; Apr
       (junk (set-tm:isdst bdtm -1))
       (tax-day (cons (car (mktime bdtm)) 0))
       (after-tax-day (gnc:timepair-later tax-day today)))
  
  (define (make-split-list account split-filter-pred)
    (reverse (filter split-filter-pred
		     (gnc:glist->list (gnc:account-get-split-list account)
				      <gnc:Split*>))))
  
  ;; returns a predicate that returns true only if a split is
  ;; between early-date and late-date
  (define (split-report-make-date-filter-predicate begin-date-tp
                                                   end-date-tp)
    (lambda (split) 
      (let ((tp
             (gnc:transaction-get-date-posted
              (gnc:split-get-parent split))))
        (and (gnc:timepair-ge-date tp begin-date-tp)
             (gnc:timepair-le-date tp end-date-tp)))))
  
  ;; This is nearly identical to, and could be shared with
  ;; display-report-list-item in report.scm. This adds warn-msg parameter
  (define (gnc:display-report-list-item item port warn-msg)
    (cond
     ((string? item) (display item port))
     ((null? item) #t)
     ((list? item) (map (lambda (item)
                          (gnc:display-report-list-item item port warn-msg))
                        item))
     (else (gnc:warn warn-msg item " is the wrong type."))))

  (define (lx-collector level action value)
    ((vector-ref levelx-collector (- level 1)) action value))

  ;; IRS asked congress to make the tax quarters the same as real quarters
  ;;   This is the year it is effective.  THIS IS A Y10K BUG!
  (define tax-qtr-real-qtr-year 10000)

  (define (tax-options-generator)
    (define options (gnc:new-options))
    (define (gnc:register-tax-option new-option)
      (gnc:register-option options new-option))

    ;; date at which to report 
    (gnc:options-add-date-interval!
     options gnc:pagename-general 
     (N_ "From") (N_ "To") "a")

    (gnc:register-tax-option
     (gnc:make-multichoice-option
      gnc:pagename-general (N_ "Alternate Period")
      "c" (N_ "Override or modify From: & To:")
      (if after-tax-day 'from-to 'last-year)
      (list (list->vector
             (list 'from-to (N_ "Use From - To") (N_ "Use From - To period")))
            (list->vector
             (list '1st-est (N_ "1st Est Tax Quarter") (N_ "Jan 1 - Mar 31")))
            (list->vector
             (list '2nd-est (N_ "2nd Est Tax Quarter") (N_ "Apr 1 - May 31")))
            (list->vector
             (list '3rd-est (N_ "3rd Est Tax Quarter") (N_ "Jun 1 - Aug 31")))
            (list->vector
             (list '4th-est (N_ "4th Est Tax Quarter") (N_ "Sep 1 - Dec 31")))
            (list->vector
             (list 'last-year (N_ "Last Year") (N_ "Last Year")))
            (list->vector
             (list '1st-last (N_ "Last Yr 1st Est Tax Qtr")
                   (N_ "Jan 1 - Mar 31, Last year")))
            (list->vector
             (list '2nd-last (N_ "Last Yr 2nd Est Tax Qtr")
                   (N_ "Apr 1 - May 31, Last year")))
            (list->vector
             (list '3rd-last (N_ "Last Yr 3rd Est Tax Qtr")
                   (N_ "Jun 1 - Aug 31, Last year")))
            (list->vector
             (list '4th-last (N_ "Last Yr 4th Est Tax Qtr")
                   (N_ "Sep 1 - Dec 31, Last year"))))))

    (gnc:register-tax-option
     (gnc:make-account-list-option
      gnc:pagename-accounts (N_ "Select Accounts (none = all)")
      "d" (N_ "Select accounts")
      (lambda () '())
      #f #t))
    
    (gnc:register-tax-option
     (gnc:make-simple-boolean-option
      gnc:pagename-display (N_ "Suppress $0.00 values")
      "f" (N_ "$0.00 valued Accounts won't be printed.") #t))

    (gnc:register-tax-option
     (gnc:make-simple-boolean-option
      gnc:pagename-display (N_ "Print Full account names")
      "g" (N_ "Print all Parent account names") #f))

    (gnc:options-set-default-section options gnc:pagename-general)

    options)

  ;; Render txf information
  (define crlf (string #\return #\newline)) ; TurboTax seems to want these
  
  (define txf-last-payer "")		; if same as current, inc txf-l-count
					; this only works if different
					; codes from the same payer are
					; grouped in the accounts list
  (define txf-l-count 0)		; count repeated N codes

  ;; stores assigned txf codes so we can check for duplicates
  (define txf-dups-alist '())

  (define (txf-payer? payer)
    (member payer (list 'current 'parent)))

  (define (gnc:account-get-txf account)
    (and (gnc:account-get-tax-related account)
	 (not (equal? (gnc:account-get-txf-code account) 'N000))))

  (define (gnc:account-get-txf-code account)
    (let ((code (gnc:account-get-tax-US-code account)))
      (string->symbol (if code code "N000"))))

  (define (gnc:get-txf-format code income?)
    (gnc:txf-get-format (if income?
                            txf-income-categories
                            txf-expense-categories)
                        code))

  (define (gnc:account-get-txf-payer-source account)
    (let ((pns (gnc:account-get-tax-US-payer-name-source account)))
      (string->symbol (if pns pns "none"))))

  ;; check for duplicate txf codes
  (define (txf-check-dups account) 
    (let* ((code (gnc:account-get-txf-code account))
	   (item (assoc-ref txf-dups-alist code))
	   (payer (gnc:account-get-txf-payer-source account)))
      (if (not (txf-payer? payer))
	  (set! txf-dups-alist (assoc-set! txf-dups-alist code
					   (if item
					       (cons account item)
					       (list account)))))))
  
  ;; Print error message for duplicate txf codes and accounts
  (define (txf-print-dups doc)
    (let ((dups
           (apply append
                  (map (lambda (x)
                         (let ((cnt (length (cdr x))))
                           (if (> cnt 1)
                               (let* ((acc (cadr x))
                                      (txf (gnc:account-get-txf acc)))
                                 (cons (string-append 
                                        "Code \"" 
                                        (symbol->string
                                         (gnc:account-get-txf-code acc))
                                        "\" has duplicates in "
                                        (number->string cnt) " accounts:")
                                       (map gnc:account-get-full-name 
                                            (cdr x))))
                               '())))
                       txf-dups-alist)))
          (text (gnc:make-html-text)))
      (if (not (null? dups))
	  (begin
            (gnc:html-document-add-object! doc text)
            (gnc:html-text-append!
             text
             (gnc:html-markup-p
              (gnc:html-markup
               "blue"
               (_ "WARNING: There are duplicate TXF codes assigned\
 to some accounts. Only TXF codes with payer sources may be repeated."))))
            (map (lambda (s)
                   (gnc:html-text-append!
                    text
                    (gnc:html-markup-p
                     (gnc:html-markup "blue" s))))
                 dups)))))

  ;; some codes require special handling
  (define (txf-special-split? code)
    (member code (list 'N521)))            ; only one for now
  
  (define (fill-clamp-sp str len)
    (string-append (substring (string-append str (make-string len #\space))
			      0 (- len 1)) " "))
  
  (define (fill-clamp str len)
    (string-append (substring (string-append str (make-string len #\space))
			      0 len)))
  
  (define (make-header-row table max-level)
    (gnc:html-table-prepend-row!
     table
     (append (list (gnc:make-html-table-header-cell/markup
		    "account-header" (_ "Account Name")))
	     (make-sub-headers max-level)
	     (list (gnc:make-html-table-header-cell/markup
		    "number-header" (_ "Total"))))))
  
  (define (make-sub-headers max-level)
    (if (<= max-level 1)
	'()
	(cons (gnc:make-html-table-header-cell/markup
	       "number-header"
	       (_ "Sub-")
	       (number->string (- max-level 1)))
	      (make-sub-headers (- max-level 1)))))
  
  (define (render-txf-account account account-value d? date x? x-date)
    (let* ((print-info (gnc:account-value-print-info account #f))
	   (value (gnc:amount->string account-value print-info))
	   (value (if (equal? "$" (substring value 0 1))
		      value
		      (string-append "$" value)))
	   (txf? (gnc:account-get-txf account)))
      (if (and txf?
	       (not (equal? value (gnc:amount->string 0 print-info))))
	  (let* ((type (gw:enum-<gnc:AccountType>-val->sym
			(gnc:account-get-type account) #f))
		 (code (gnc:account-get-txf-code account))
		 (date-str (if date
			       (strftime "%m/%d/%Y" (localtime (car date)))
			       #f))
		 (x-date-str (if x-date
				 (strftime "%m/%d/%Y" (localtime (car x-date)))
				 #f))
		 ;; Only formats 1,3 implemented now! Others are treated as 1.
		 (format (gnc:get-txf-format code (eq? type 'income)))
		 (payer-src (gnc:account-get-txf-payer-source account))
		 (account-name (if (eq? payer-src 'parent)
				   (gnc:account-get-name
				    (gnc:group-get-parent
				     (gnc:account-get-parent account)))
				   (gnc:account-get-name account))) 
		 (action (if (eq? type 'income)
			     (case (string->symbol code)
			       ((N286 N488) "ReinvD")
			       (else "Income"))
			     "Expense"))
		 (category-key (if (eq? type 'income)
				   (gnc:txf-get-category-key 
				    txf-income-categories code)
				   (gnc:txf-get-category-key
				    txf-expense-categories code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (string-length value))
				  " code)))
		 (value-name (if (equal? "ReinvD" action)
				 (string-append 
				  (substring value 1 (st